/*
 * joystick_commander
 * 2018.04.19 YMG
 *
 * ロボ研にあるPS3とLogicoolの無線コントローラから
 * move_baseへの速度指令/cmd_velをPublishするノード
 *
 * <usage>
 * $ roscore
 * $ rosrun joystick_commander joystick_commander /dev/input/js0
 * ($ rosrun joystick_commander joystick_commander /dev/input/js0 _debug_mode:=true)
 */

#include <iostream>

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>

#include <linux/input.h>
#include <linux/joystick.h>

#include "ros/ros.h"
#include "geometry_msgs/Twist.h"

#include "axbtnmap.h"

#define NAME_LENGTH 128

// rad を deg に換算
inline const double rad2deg(const double a_angle_rad)
{/*{{{*/
	return a_angle_rad / M_PI * 180.0;
}/*}}}*/

// deg を rad に換算
inline const double deg2rad(const double a_angle_deg)
{/*{{{*/
	return a_angle_deg / 180.0 * M_PI;
}/*}}}*/

// 操作モード
// 0：Logitech Logitech Attack 3 用
// 1：Sony PLAYSTATION(R)3 Controller 用
// 2：Sony PLAYSTATION(R)3 Controller 用 軸数が少なく検出されてしまう場合に対応するため
const	int JS_TYPE_UNKNOWN  =-1;
const	int JS_TYPE_LOGETECH = 0;
const	int JS_TYPE_PS3      = 1;
const	int JS_TYPE_PS3_B    = 2;

const int AXIS_TRANS   = 0;
const int AXIS_ROT     = 1;
const int BTN_TRANS_UP = 2;
const int BTN_TRANS_DN = 3;
const int BTN_ROT_UP   = 4;
const int BTN_ROT_DN   = 5;

std::vector< std::vector<int> > js_info_list {
	{ 1,  2,  4,  6,  5,  7 },   // Logitech
	{ 1,  2, 10,  8, 11,  9 },   // PS3
	{ 1,  3,  6,  4,  7,  5 }    // PS3 few axis
};


const double AXIS_MAX_VALUE = 32767.0;

double trans_vel_mps = 0.0, angular_vel_radps = 0.0;
double input_trans_vel_ratio = 0.0, input_angular_vel_ratio = 0.0;
double max_trans_vel_mps = 0.3;  // 平身速度の最大値の初期値
double max_angular_vel_radps = deg2rad(15.0); // 回転角速度の最大値の初期値

bool is_ready = false;

// 速度を表示
void showInfo ( const ros::TimerEvent& event )
{/*{{{*/
	if ( ! is_ready ) return;

	// 2行分にカーソルを移動させ、その行の出力を消去
	printf( "\x1b\[1A" );
	printf( "\x1b\[K" );
	printf( "\x1b\[1A" );
	printf( "\x1b\[K" );
	printf( ">> max trans vel : %2.4f,   max rot vel : %2.4f\n", max_trans_vel_mps, rad2deg(max_angular_vel_radps) );
	printf( ">> vel : %2.4f,   rot vel : %2.4f\n", trans_vel_mps, rad2deg(angular_vel_radps) );
}/*}}}*/

ros::Publisher cmd_publisher;
void publishVelocity ( const ros::TimerEvent& event )
{/*{{{*/
	if ( ! is_ready ) return;

	// 速度の計算
	trans_vel_mps = input_trans_vel_ratio * max_trans_vel_mps;
	angular_vel_radps = input_angular_vel_ratio * max_angular_vel_radps;
	// 後進は速度を遅くする
	if (trans_vel_mps < 0) {
		trans_vel_mps *= 0.3;
		angular_vel_radps *= 0.5;
	}

	geometry_msgs::Twist cmd_vel_msg;
	cmd_vel_msg.linear.x = trans_vel_mps;
	cmd_vel_msg.linear.y = 0.0;
	cmd_vel_msg.linear.z = 0.0;
	cmd_vel_msg.angular.x = 0.0;
	cmd_vel_msg.angular.y = 0.0;
	cmd_vel_msg.angular.z = angular_vel_radps;
	cmd_publisher.publish(cmd_vel_msg);
}/*}}}*/

int main(int argc, char **argv) {
	// 引数が正しくない場合は使い方を表示
	if (argc != 2) {
		printf ( "USAGE:\n" );
		printf ( "    $ rosrun joystick_commander joystick_commander <dev_file_path>\n" );
		printf ( "e.g.) rosrun joystick_commander joystick_commander /dev/input/js0\n" );
		return -1;
	}

	ros::init(argc, argv, "joystick_commander");
	ros::NodeHandle nh, private_nh("~");

	bool debug_mode;
	double control_frequency;
	double show_frequency;
	private_nh.param( "debug_mode", debug_mode, false );
	private_nh.param( "control_frequency", control_frequency, 100.0 );   // コマンド送信周期100Hz
	private_nh.param( "show_frequency", show_frequency, 20.0 );   // コンソール表示周期20Hz

	cmd_publisher = nh.advertise<geometry_msgs::Twist>("cmd_vel", 10);
	ros::Timer publish_velocity_timer = nh.createTimer( ros::Duration( 1/control_frequency ), publishVelocity );
	ros::Timer show_info_timer = nh.createTimer( ros::Duration( 1/show_frequency ), showInfo );

	// 最大最小速度
	const double max_limit_trans_vel_mps = 1.1;  // 1.1 m/s = 3.96 km/h
	const double max_limit_angular_vel_radps = deg2rad(45.0);
	const double min_limit_trans_vel_mps = 0.1;
	const double min_limit_angular_vel_radps = deg2rad(5.0);

	// ボタン入力による速度調整幅
	const double limit_step_trans_vel_mps = 0.1;
	const double limit_step_angular_vel_radps = deg2rad(5.0);

	// ジョイスティック関連
	int fd;
	unsigned char num_of_axes = 2;
	unsigned char num_of_buttons = 2;
	int version = 0x000800;
	char name[NAME_LENGTH] = "Unknown";
	uint16_t btnmap[BTNMAP_SIZE];
	uint8_t axmap[AXMAP_SIZE];
	int btnmapok = 1;

	if ((fd = open(argv[argc - 1], O_RDONLY)) < 0) {
		perror("ERROR: Can't connect to the joystick!! ");
		return -1;
	}

	ioctl(fd, JSIOCGVERSION, &version);
	ioctl(fd, JSIOCGAXES, &num_of_axes);
	ioctl(fd, JSIOCGBUTTONS, &num_of_buttons);
	ioctl(fd, JSIOCGNAME(NAME_LENGTH), name);

	// nonblocking mode
	fcntl(fd, F_SETFL, O_NONBLOCK);

	getaxmap(fd, axmap);
	getbtnmap(fd, btnmap);

	printf("Driver version: %d.%d.%d\n", version >> 16, (version >> 8) & 0xff, version & 0xff);

	/* Determine whether the button map is usable. */
	for (int i = 0; btnmapok && i < num_of_buttons; ++i) {
		if (btnmap[i] < BTN_MISC || btnmap[i] > KEY_MAX) {
			btnmapok = 0;
			break;
		}
	}

	/* btnmap out of range for names */
	if (!btnmapok) printf("INFO: Not fully compatible with your kernel. Unable to retrieve button map!!");

	printf("Joystick (%s) has %d axes and %d buttons.\n", name, num_of_axes, num_of_buttons);

	// ジョイスティック名による操作モード変更
	int joystick_type = JS_TYPE_UNKNOWN;
	if (strncmp(name, "Logitech Logitech Cordless RumblePad 2", 38) == 0) {
		joystick_type = JS_TYPE_LOGETECH;
	}
	else if (strncmp(name, "Sony PLAYSTATION(R)3 Controller", 31) == 0) {
		joystick_type = JS_TYPE_PS3;
		if ( num_of_axes == 6 ) joystick_type = JS_TYPE_PS3_B;
	}
	else {
		ROS_INFO("unknown joystick found");
		return -1;
	}
	std::vector<int> js_info = js_info_list[joystick_type];

	struct js_event read_event;
	std::vector<int> axis_states(num_of_axes, 0);
	std::vector<char> button_states(num_of_buttons, 0);

	printf("\n");
	printf("Press PS button to start.\n");

	// メインループ
	while ( ros::ok() ) {
		ros::Duration(0.005).sleep();
		ros::spinOnce();

		// 入力イベントを読んだかどうか
		bool is_read_event = false;
		if (read(fd, &read_event, sizeof(read_event)) == sizeof(read_event)) {
			is_read_event = true;
		}
		else if (errno != EAGAIN) {  // エラー判定
			perror("ERROR: Read joystick error!! ");
			return -1;
		}

		// 入力イベントを読まなかったら処理を行わない
		if ( ! is_read_event ) continue;

		// 起動後2秒間はジョイスティックの入力を無視する
		// ジョイスティックを傾けた状態で、USBケーブルを抜いてプロセスが落ちると、次回の起動直後にmax値が代入される
		static ros::Time start_time = ros::Time::now();
		if ( ros::Time::now() - start_time < ros::Duration(2.0) ) {
			continue;
		}

		// 送信コマンドのパラメータ設定
		int target_event = -1;
		switch (read_event.type & ~JS_EVENT_INIT) {
			case JS_EVENT_AXIS:
				axis_states[read_event.number] = read_event.value;
				break;
			case JS_EVENT_BUTTON:

				// ボタンを押したらスタート
				is_ready = true;

				button_states[read_event.number] = read_event.value;

				// increase max trans vel
				target_event = js_info[BTN_TRANS_UP];
				if (read_event.number == target_event && button_states[target_event]) {
					max_trans_vel_mps = std::min( max_trans_vel_mps + limit_step_trans_vel_mps, max_limit_trans_vel_mps );
				}
				// decrease max trans vel
				target_event = js_info[BTN_TRANS_DN];
				if (read_event.number == target_event && button_states[target_event]) {
					max_trans_vel_mps = std::max( max_trans_vel_mps - limit_step_trans_vel_mps, min_limit_trans_vel_mps );
				}
				// increase max angular vel
				target_event = js_info[BTN_ROT_UP];
				if (read_event.number == target_event && button_states[target_event]) {
					max_angular_vel_radps = std::min( max_angular_vel_radps + limit_step_angular_vel_radps, max_limit_angular_vel_radps );
				}
				// decrease max angular vel
				target_event = js_info[BTN_ROT_DN];
				if (read_event.number == target_event && button_states[target_event]) {
					max_angular_vel_radps = std::max( max_angular_vel_radps - limit_step_angular_vel_radps, min_limit_angular_vel_radps );
				}

				break; // end case JS_EVENT_BUTTON
		}

		int trans_axis_value = axis_states[ js_info[AXIS_TRANS] ];
		int rot_axis_value = axis_states[ js_info[AXIS_ROT] ];

		// if ( joystick_type == JS_TYPE_LOGETECH ) {
		// 	rot_axis_value = rot_axis_value * 2 + AXIS_MAX_VALUE;
		// 	// Logitec のjoystickは入力なしで正確に0とならないので、不感帯を設定
		// 	static int dead_zone_value = 8000;
		// 	if ( -dead_zone_value < trans_axis_value && trans_axis_value < dead_zone_value ) trans_axis_value = 0;
		// 	if ( -dead_zone_value < rot_axis_value && rot_axis_value < dead_zone_value ) rot_axis_value = 0;
		// }

		input_trans_vel_ratio = trans_axis_value / -AXIS_MAX_VALUE;
		input_angular_vel_ratio = rot_axis_value / -AXIS_MAX_VALUE;

		if ( debug_mode ) {
			// ジョイスティック状態表示
			if (num_of_axes) {
				printf("Axes: ");
				for (int i = 0; i < num_of_axes; ++i) {
					printf("%2d:%6d ", i, axis_states[i]);
				}
			}
			printf("\n");
			if (num_of_buttons) {
				printf("Buttons: ");
				for (int i = 0; i < num_of_buttons; ++i) {
					printf("%2d:%s ", i, button_states[i] ? "on " : "off");
				}
			}
			printf("\n");
		}

	} // while ( ros::ok() )

return 0;
}
